#!/usr/bin/env python
"""Module with static website supporting infrastructure."""
from __future__ import print_function

import logging

import awacs.s3
from awacs.aws import Allow, AWSPrincipal, Policy, Statement
from troposphere import AccountId, Join, Output, cognito, s3

from runway.cfngin.blueprints.base import Blueprint

LOGGER = logging.getLogger(__name__)


class Dependencies(Blueprint):
    """Stacker blueprint for creating static website buckets."""

    VARIABLES = {
        'AuthAtEdge': {
            'type': bool,
            'default': False,
            'description': 'Utilizing Authorization @ Edge'
        },
        'CreateUserPool': {
            'type': bool,
            'default': False,
            'description': 'Whether a User Pool should be created for the project'
        },
        'UserPoolId': {
            'type': str,
            'default': '',
            'description': 'User Pool ID for Authorization @ Edge'
        },
        'OAuthScopes': {
            'type': list,
            'default': [
                'phone',
                'email',
                'profile',
                'openid',
                'aws.cognito.signin.user.admin'
            ],
            'description': 'The allowed scopes for OAuth validation'
        }
    }

    def create_template(self):
        """Create template (main function called by Stacker)."""
        template = self.template
        variables = self.get_variables()
        template.set_version('2010-09-09')
        template.set_description('Static Website - Dependencies')

        # Resources
        awslogbucket = template.add_resource(
            s3.Bucket(
                'AWSLogBucket',
                AccessControl=s3.Private,
                VersioningConfiguration=s3.VersioningConfiguration(
                    Status='Enabled'
                )
            )
        )
        template.add_output(Output(
            'AWSLogBucketName',
            Description='Name of bucket storing AWS logs',
            Value=awslogbucket.ref()
        ))

        template.add_resource(
            s3.BucketPolicy(
                'AllowAWSLogWriting',
                Bucket=awslogbucket.ref(),
                PolicyDocument=Policy(
                    Version='2012-10-17',
                    Statement=[
                        Statement(
                            Action=[awacs.s3.PutObject],
                            Effect=Allow,
                            Principal=AWSPrincipal(Join(':',
                                                        ['arn:aws:iam:',
                                                         AccountId,
                                                         'root'])),
                            Resource=[
                                Join('', ['arn:aws:s3:::',
                                          awslogbucket.ref(),
                                          '/*'])
                            ]
                        )
                    ]
                )
            )
        )
        artifacts = template.add_resource(
            s3.Bucket(
                'Artifacts',
                AccessControl=s3.Private,
                LifecycleConfiguration=s3.LifecycleConfiguration(
                    Rules=[
                        s3.LifecycleRule(
                            NoncurrentVersionExpirationInDays=90,
                            Status='Enabled'
                        )
                    ]
                ),
                VersioningConfiguration=s3.VersioningConfiguration(
                    Status='Enabled'
                )
            )
        )
        template.add_output(Output(
            'ArtifactsBucketName',
            Description='Name of bucket storing artifacts',
            Value=artifacts.ref()
        ))

        if variables['AuthAtEdge']:
            callbacks = self.context.hook_data['aae_callback_url_retriever']['callback_urls']

            user_pool_id = variables['UserPoolId']

            if variables['CreateUserPool']:
                user_pool = template.add_resource(
                    cognito.UserPool("AuthAtEdgeUserPool")
                )

                user_pool_id = user_pool.ref()

                template.add_output(Output(
                    'AuthAtEdgeUserPoolId',
                    Description='Cognito User Pool App Client for Auth @ Edge',
                    Value=user_pool_id
                ))

            client = template.add_resource(
                cognito.UserPoolClient(
                    "AuthAtEdgeClient",
                    AllowedOAuthFlows=['code'],
                    CallbackURLs=callbacks,
                    UserPoolId=user_pool_id,
                    AllowedOAuthScopes=variables['OAuthScopes']
                )
            )

            template.add_output(Output(
                'AuthAtEdgeClient',
                Description='Cognito User Pool App Client for Auth @ Edge',
                Value=client.ref()
            ))


# Helper section to enable easy blueprint -> template generation
# (just run `python <thisfile>` to output the json)
if __name__ == "__main__":
    from runway.cfngin.context import Context
    print(Dependencies('test', Context({"namespace": "test"}), None).to_json())



"""CFNgin info action."""
import logging

from .. import exceptions
from .base import BaseAction

LOGGER = logging.getLogger(__name__)


class Action(BaseAction):  # pylint: disable=abstract-method
    """Get information on CloudFormation stacks.

    Displays the outputs for the set of CloudFormation stacks.

    """

    def run(self, **kwargs):
        """Get information on CloudFormation stacks."""
        LOGGER.info('Outputs for stacks: %s', self.context.get_fqn())
        if not self.context.get_stacks():
            LOGGER.warning('WARNING: No stacks detected (error in config?)')
        for stack in self.context.get_stacks():
            provider = self.build_provider(stack)

            try:
                provider_stack = provider.get_stack(stack.fqn)
            except exceptions.StackDoesNotExist:
                LOGGER.info('Stack "%s" does not exist.', stack.fqn,)
                continue

            LOGGER.info('%s:', stack.fqn)
            if 'Outputs' in provider_stack:
                for output in provider_stack['Outputs']:
                    LOGGER.info(
                        '\t%s: %s',
                        output['OutputKey'],
                        output['OutputValue']
                    )



"""Tests for the CDK module."""
import os
import glob

from send2trash import send2trash

from integration_tests.integration_test import IntegrationTest
from integration_tests.util import (copy_file, copy_dir, import_tests,
                                    execute_tests)


class CDK(IntegrationTest):
    """Test CDK based module scenarios"""
    base_dir = os.path.abspath(os.path.dirname(__file__))
    fixtures_dir = os.path.join(base_dir, 'fixtures')
    multiple_stacks_dir = os.path.join(fixtures_dir, 'multiple-stacks-app.cdk')
    tests_dir = os.path.join(base_dir, 'tests')

    cdk_test_dir = os.path.join(base_dir, 'cdk_test')

    def copy_fixture(self, name='multiple-stacks-app.cdk'):
        """Copy fixture files for test"""
        copy_dir(
            os.path.join(self.fixtures_dir, name),
            os.path.join(self.cdk_test_dir, name)
        )

    def copy_runway(self, template):
        """Copy runway template to proper directory."""
        template_file = os.path.join(self.fixtures_dir, 'runway-{}.yml'.format(template))
        copy_file(template_file, os.path.join(self.cdk_test_dir, 'runway.yml'))

    def run(self):
        """Find all tests and run them."""
        import_tests(self.logger, self.tests_dir, 'test_*')
        tests = [test(self.logger) for test in CDK.__subclasses__()]
        if not tests:
            raise Exception('No tests were found.')
        self.logger.debug('FOUND TESTS: %s', tests)
        self.set_environment('dev')
        self.set_env_var('PIPENV_VENV_IN_PROJECT', '1')
        err_count = execute_tests(tests, self.logger)
        assert err_count == 0  # assert that all subtests were successful
        return err_count

    def clean(self):
        """Clean up CDK module directory."""
        file_types = ('*.yaml', '*.yml')
        templates = []
        for file_type in file_types:
            templates.extend(glob.glob(os.path.join(self.cdk_test_dir, file_type)))
        for template in templates:
            if os.path.isfile(template):
                self.logger.debug('send2trash: "%s"', template)
                send2trash(template)
        folders = ['multiple-stacks-app.cdk', 'decline-deploy-app.cdk']
        for folder in folders:
            folder_path = os.path.join(self.cdk_test_dir, folder)
            if os.path.isdir(folder_path):
                self.logger.debug('send2trash: "%s"', folder_path)
                send2trash(folder_path)

    def delete_venv(self, module_directory):
        """Delete pipenv venv before running destroy."""
        folder_path = os.path.join(self.cdk_test_dir,
                                   f'{module_directory}/.venv')
        if os.path.isdir(folder_path):
            self.logger.debug('send2trash: "%s"', folder_path)
            send2trash(folder_path)

    def teardown(self):
        """Teardown resources create during init."""
        self.unset_env_var('PIPENV_VENV_IN_PROJECT')
        self.clean()



"""Base class for CFNgin hooks."""
import logging

from troposphere import Tags
from six.moves import UserDict

from runway.util import MutableMap

from ..actions import build
from ..exceptions import StackFailed
from ..plan import COLOR_CODES
from ..stack import Stack
from ..status import COMPLETE, FAILED, PENDING, SKIPPED, SUBMITTED

LOGGER = logging.getLogger(__name__)


class Hook(object):
    """Base class for hooks.

    Not all hooks need to be classes and not all classes need to be hooks.

    Attributes:
        args (MutableMap): Keyword arguments passed to the hook, loaded into
            a MutableMap object.
        blueprint (Optional[Blueprint]): Blueprint generated by the hook if
            it will be deploying a stack.
        context (Context): Context instance. (passed in by CFNgin)
        provider (BaseProvider): Provider instance. (passed in by CFNgin)
        stack (Optional[Stack]): Stack object if the hook deploys a stack.
        stack_name (str): Name of the stack created by the hook if a stack is
            to be created.

    """

    def __init__(self, context, provider, **kwargs):
        """Instantiate class.

        Args:
            context (:class:`runway.cfngin.context.Context`): Context instance.
                (passed in by CFNgin)
            provider (:class:`runway.cfngin.providers.base.BaseProvider`):
                Provider instance. (passed in by CFNgin)

        """
        kwargs.setdefault('tags', {})

        self.args = MutableMap(**kwargs)
        self.args.tags.update(context.tags)
        self.blueprint = None
        self.context = context
        self.provider = provider
        self.stack = None
        self.stack_name = 'stack'
        self._deploy_action = HookBuildAction(self.context, self.provider)
        self._destroy_action = HookDestroyAction(self.context, self.provider)

    @property
    def tags(self):
        """Return tags that should be applied to any resource being created.

        Returns:
            troposphere.Tags

        """
        return Tags(**dict(self.context.tags, **self.args.tags.data))

    def generate_stack(self, **kwargs):
        """Create a CFNgin Stack object.

        Returns:
            Stack

        """
        definition = HookStackDefinition(name=self.stack_name,
                                         tags=self.args.tags.data,
                                         **kwargs)
        stack = Stack(definition, self.context)
        stack._blueprint = self.blueprint  # pylint: disable=protected-access
        return stack

    def get_template_description(self, suffix=None):
        """Generate a template description.

        Args:
            suffix (Optional[str]): Suffix to append to the end of a
                CloudFormation template description.

        Returns:
            str: CloudFormation template description.

        """
        template = 'Automatically generated by {}'
        if suffix:
            template += ' - {}'
            return template.format(self.__class__.__module__, suffix)
        return template.format(self.__class__.__module__)

    def deploy_stack(self, stack=None, wait=False):
        """Deploy a stack.

        Args:
            stack (Optional[Stack]): A stack to act on.
            wait (bool): Wither to wait for the stack to complete before
                returning.

        Returns:
            Status: Ending status of the stack.

        """
        return self._run_stack_action(action=self._deploy_action,
                                      stack=stack,
                                      wait=wait)

    def destroy_stack(self, stack=None, wait=None):
        """Destroy a stack.

        Args:
            stack (Optional[Stack]): A stack to act on.
            wait (bool): Wither to wait for the stack to complete before
                returning.

        Returns:
            Status: Ending status of the stack.

        """
        return self._run_stack_action(action=self._destroy_action,
                                      stack=stack,
                                      wait=wait)

    def post_deploy(self):
        """Run during the **post_deploy** stage."""
        raise NotImplementedError

    def post_destroy(self):
        """Run during the **post_destroy** stage."""
        raise NotImplementedError

    def pre_deploy(self):
        """Run during the **pre_deploy** stage."""
        raise NotImplementedError

    def pre_destroy(self):
        """Run during the **pre_destroy** stage."""
        raise NotImplementedError

    @staticmethod
    def _log_stack(stack, status):
        """Log stack status. Mimics normal stack deployment.

        Args:
            stack (:class:`runway.cfngin.stack.Stack`): The stack being logged.
            status (:class:`runway.cfngin.status.Status`) The status being
                logged.

        """
        msg = "%s: %s" % (stack.name, status.name)
        if status.reason:
            msg += " (%s)" % (status.reason)
        color_code = COLOR_CODES.get(status.code, 37)
        LOGGER.info(msg, extra={"color": color_code})

    def _run_stack_action(self, action, stack=None, wait=None):
        """Run a CFNgin hook modified for use in hooks.

        Args:
            action (BaseAction): Action to be taken against a stack.
            stack (Optional[Stack]): A stack to act on.
            wait (bool): Wither to wait for the stack to complete before
                returning.

        Returns:
            Status: Ending status of the stack.

        """
        stack = stack or self.stack
        status = action.run(stack=stack, status=PENDING)
        self._log_stack(stack, status)

        if wait and status != SKIPPED:
            status = self._wait_for_stack(action=action, stack=stack,
                                          last_status=status)
        return status

    def _wait_for_stack(self, action, last_status=None, stack=None,
                        till_reason=None):
        """Wait for a CloudFormation stack to complete.

        Args:
            action (BaseAction): Action to be taken against a stack.
            last_status (Optional[Status]): The last status of the stack.
            stack (Optional[Stack]): A stack that has been acted upon.
            till_reason (Optional[str]): Status string to wait for before
                returning. ``COMPLETE`` or ``FAILED`` status will return
                before this condition if found.

        Returns:
            Status: Ending status of the stack.

        Raises:
            StackFailed: Stack is in a failed state.

        """
        status = last_status or SUBMITTED
        stack = stack or self.stack

        while True:
            if status in (COMPLETE, FAILED):
                break
            if (till_reason and status.reason) and status.reason == till_reason:
                break
            if last_status and last_status.reason != status.reason:
                # log status changes like rollback
                self._log_stack(stack, status)
                last_status = status
            LOGGER.debug('Waiting for stack to complete...')
            status = action.run(stack=stack, status=status)

        self._log_stack(stack, status)
        if status == FAILED:
            raise StackFailed(stack_name=stack.fqn,
                              status_reason=status.reason)
        return status


class HookBuildAction(build.Action):
    """Build action that can be used from hooks."""

    def __init__(self, context, provider):
        """Instantiate class.

        Args:
            context (:class:`runway.cfngin.context.Context`): The context
                for the current run.

        """
        super(HookBuildAction, self).__init__(context)
        self._provider = provider

    @property
    def provider(self):
        """Override the inherited property to return the local provider."""
        return self._provider

    def build_provider(self, stack):
        """Override the inherited method to always return local provider."""
        return self._provider

    def run(self, **kwargs):
        """Run the action for one stack."""
        return self._launch_stack(**kwargs)


# the build action has logic to destroy stacks so we can just extend the
# HookBuildAction and change `run` in use the `_destroy_stack` method instead
class HookDestroyAction(HookBuildAction):
    """Destroy action that can be used from hooks."""

    def run(self, **kwargs):
        """Run the action for one stack."""
        return self._destroy_stack(**kwargs)


# TODO remove multiple inheritance when droping python 2 support
class HookStackDefinition(UserDict, object):
    """Stack definition for use in hooks to avoid cyclic imports."""

    def __init__(self, name, **kwargs):
        """Instantiate class."""
        values = {
            'class_path': None,
            'enabled': True,
            'in_progress_behavior': None,
            'name': name,
            'locked': False,
            'profile': None,
            'protected': False,
            'region': None,
            'required_by': None,
            'requires': None,
            'stack_name': None,
            'stack_policy_path': None,
            'tags': None,
            'template_path': None,
            'termination_protection': False,
            'variables': None
        }
        values.update(kwargs)
        super(HookStackDefinition, self).__init__(**values)

    def __getattr__(self, key):
        """Implement dot notation."""
        try:
            return self[key]
        except KeyError:
            raise AttributeError(key)



"""Tests for runway.cfngin.hooks.base."""
# pylint: disable=no-self-use,too-few-public-methods
import logging

import pytest
from mock import MagicMock, call, patch

from runway.cfngin.exceptions import StackFailed
from runway.cfngin.hooks.base import (Hook, HookBuildAction, HookDestroyAction,
                                      HookStackDefinition)
from runway.cfngin.status import (COMPLETE, FAILED, SKIPPED, SUBMITTED,
                                  CompleteStatus, SubmittedStatus)

COMPLETE_W_REASON = CompleteStatus('test successful')


class TestHook(object):
    """Tests for runway.cfngin.hooks.base.Hook."""

    def test_attributes(self, cfngin_context):
        """Test attributes set during __init__."""
        provider = MagicMock()
        args = {
            'key': 'val'
        }
        result = Hook(cfngin_context, provider, **args)

        assert result.args.key == 'val'
        assert result.args.tags == {}
        assert not result.blueprint
        assert result.context == cfngin_context
        assert result.provider == provider
        assert not result.stack
        assert result.stack_name == 'stack'

    def test_tags(self, cfngin_context):
        """Test tags property."""
        cfngin_context.config['tags'] = {'context_tag': 'val'}

        hook = Hook(cfngin_context, MagicMock(), **{'tags': {'arg_tag': 'val'}})

        assert hook.tags.to_dict() == [{'Key': 'arg_tag', 'Value': 'val'},
                                       {'Key': 'context_tag', 'Value': 'val'}]

    def test_get_template_description(self, cfngin_context):
        """Test for get_template_description."""
        hook = Hook(cfngin_context, MagicMock())

        result = hook.get_template_description()

        assert result == 'Automatically generated by {}'.format(
            hook.__class__.__module__
        )
        assert hook.get_template_description('suffix').endswith(' - suffix')

    @patch('runway.cfngin.hooks.base.HookBuildAction.run',
           MagicMock(return_value=COMPLETE))
    def test_deploy_stack(self, cfngin_context, caplog):
        """Test for deploy_stack."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock()
        stack.name = 'test-stack'

        with caplog.at_level(logging.INFO, logger='runway.cfngin.hooks.base'):
            assert hook.deploy_stack(stack=stack, wait=False) == COMPLETE

        assert caplog.records[0].message == '%s: %s' % (stack.name,
                                                        COMPLETE.name)

    @patch('runway.cfngin.hooks.base.HookBuildAction.run',
           MagicMock(side_effect=[SUBMITTED, COMPLETE]))
    def test_deploy_stack_wait(self, cfngin_context, caplog):
        """Test for deploy_stack with wait."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock()
        stack.name = 'test-stack'

        with caplog.at_level(logging.DEBUG, logger='runway.cfngin.hooks.base'):
            assert hook.deploy_stack(stack=stack, wait=True) == COMPLETE

        assert caplog.records[0].message == '%s: %s' % (stack.name,
                                                        SUBMITTED.name)
        assert caplog.records[1].message == 'Waiting for stack to complete...'
        assert caplog.records[2].message == '%s: %s' % (stack.name,
                                                        COMPLETE.name)

    @patch('runway.cfngin.hooks.base.HookBuildAction.run',
           MagicMock(side_effect=[SKIPPED]))
    def test_deploy_stack_wait_skipped(self, cfngin_context, caplog):
        """Test for deploy_stack with wait and skip."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock()
        stack.name = 'test-stack'

        with caplog.at_level(logging.INFO, logger='runway.cfngin.hooks.base'):
            assert hook.deploy_stack(stack=stack, wait=True) == SKIPPED

        assert caplog.records[0].message == '%s: %s' % (stack.name,
                                                        SKIPPED.name)

    @patch('runway.cfngin.hooks.base.HookBuildAction.run',
           MagicMock(side_effect=[FAILED]))
    def test_deploy_stack_wait_failed(self, cfngin_context):
        """Test for deploy_stack with wait and skip."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock()
        stack.name = 'test-stack'

        with pytest.raises(StackFailed):
            assert hook.deploy_stack(stack=stack, wait=True) == FAILED

    @patch('runway.cfngin.hooks.base.HookDestroyAction.run',
           MagicMock(side_effect=[SUBMITTED, COMPLETE_W_REASON]))
    def test_destroy_stack(self, cfngin_context, caplog):
        """Test for destroy_stack with wait."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock()
        stack.name = 'test-stack'

        with caplog.at_level(logging.DEBUG, logger='runway.cfngin.hooks.base'):
            assert hook.destroy_stack(stack=stack, wait=True) == COMPLETE

        assert caplog.records[0].message == '%s: %s' % (stack.name,
                                                        SUBMITTED.name)
        assert caplog.records[1].message == 'Waiting for stack to complete...'
        assert caplog.records[2].message == '%s: %s (%s)' % (
            stack.name, COMPLETE_W_REASON.name, COMPLETE_W_REASON.reason
        )

    def test_wait_for_stack_till_reason(self, cfngin_context):
        """Test _wait_for_stack till_reason option."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock(fqn='test-stack', name='stack')
        action = MagicMock()
        action.run.side_effect = [SUBMITTED,
                                  SubmittedStatus('not yet'),
                                  SubmittedStatus('catch'),
                                  COMPLETE]

        result = hook._wait_for_stack(action,  # pylint: disable=protected-access
                                      stack=stack,
                                      till_reason='catch')
        assert result == SUBMITTED
        assert result.reason == 'catch'

    def test_wait_for_stack_log_change(self, cfngin_context, monkeypatch):
        """Test _wait_for_stack log status change."""
        hook = Hook(cfngin_context, MagicMock())
        stack = MagicMock(fqn='test-stack', name='stack')
        new_status = SubmittedStatus('new')
        action = MagicMock()
        action.run.side_effect = [new_status, COMPLETE]
        mock_log = MagicMock()

        monkeypatch.setattr(hook, '_log_stack', mock_log)

        hook._wait_for_stack(action,  # pylint: disable=protected-access
                             last_status=SubmittedStatus('original'),
                             stack=stack,
                             till_reason='catch')

        mock_log.assert_has_calls([call(stack, new_status),
                                   call(stack, COMPLETE)])
        assert mock_log.call_count == 2

    def test_post_deploy(self, cfngin_context):
        """Test post_deploy."""
        hook = Hook(cfngin_context, MagicMock())

        with pytest.raises(NotImplementedError):
            hook.post_deploy()

    def test_post_destroy(self, cfngin_context):
        """Test post_destroy."""
        hook = Hook(cfngin_context, MagicMock())

        with pytest.raises(NotImplementedError):
            hook.post_destroy()

    def test_pre_deploy(self, cfngin_context):
        """Test pre_deploy."""
        hook = Hook(cfngin_context, MagicMock())

        with pytest.raises(NotImplementedError):
            hook.pre_deploy()

    def test_pre_destroy(self, cfngin_context):
        """Test pre_destroy."""
        hook = Hook(cfngin_context, MagicMock())

        with pytest.raises(NotImplementedError):
            hook.pre_destroy()


class TestHookBuildAction(object):
    """Tests for runway.cfngin.hooks.base.HookBuildAction."""

    def test_provider(self, cfngin_context):
        """Test provider property."""
        provider = MagicMock()
        obj = HookBuildAction(cfngin_context, provider)

        assert obj.provider == provider

    def test_build_provider(self, cfngin_context):
        """Test build_provider."""
        provider = MagicMock()
        obj = HookBuildAction(cfngin_context, provider)

        assert obj.build_provider(None) == provider

    def test_run(self, cfngin_context, monkeypatch):
        """Test run."""
        obj = HookBuildAction(cfngin_context, MagicMock())
        monkeypatch.setattr(obj, '_launch_stack', lambda: 'success')

        assert obj.run() == 'success'


class TestHookDestroyAction(object):
    """Tests for runway.cfngin.hooks.base.HookDestroyAction."""

    def test_run(self, cfngin_context, monkeypatch):
        """Test run."""
        obj = HookDestroyAction(cfngin_context, MagicMock())
        monkeypatch.setattr(obj, '_destroy_stack', lambda: 'success')

        assert obj.run() == 'success'


class TestHookStackDefinition(object):
    """Tests for runway.cfngin.hooks.base.HookStackDefinition."""

    def test_getattr(self):
        """Test __getattr__."""
        obj = HookStackDefinition('test')

        assert obj.name == 'test'

        with pytest.raises(AttributeError):
            assert obj.invalid



"""Tests for runway.cfngin.hooks.acm."""
# pylint: disable=no-self-use,protected-access,unused-argument
from datetime import datetime

import boto3
import pytest
from botocore.exceptions import ClientError
from botocore.stub import ANY, Stubber
from mock import MagicMock
from troposphere.certificatemanager import Certificate as CertificateResource

from runway.cfngin.exceptions import (StackDoesNotExist, StackFailed,
                                      StackUpdateBadStatus)
from runway.cfngin.hooks.acm import Certificate
from runway.cfngin.status import FAILED, NO_CHANGE, SubmittedStatus
from runway.util import MutableMap

STATUS = MutableMap(**{
    'failed': FAILED,
    'new': SubmittedStatus('creating new stack'),
    'no': NO_CHANGE,
    'recreate': SubmittedStatus('destroying stack for re-creation'),
    'update': SubmittedStatus('updating existing stack')
})


def check_bool_is_true(val):
    """Check if a value is a true bool."""
    if val and isinstance(val, bool):
        return True
    raise ValueError('Value should be "True"; got {}'.format(val))


def check_bool_is_false(val):
    """Check if a value is a false bool."""
    if not val and isinstance(val, bool):
        return True
    raise ValueError('Value should be "False"; got {}'.format(val))


def gen_certificate(**kwargs):
    """Generate a response to describe_certificate."""
    data = {
        'CertificateArn': kwargs.pop('CertificateArn'),
        'DomainName': 'place_holder_domain_name',
        # 'SubjectAlternativeNames': [],
        # 'DomainValidationOptions': []
    }
    data.update(kwargs)
    return {'Certificate': data}


def gen_change(record_set, action='CREATE'):
    """Generate expected change."""
    return {
        'Action': action,
        'ResourceRecordSet': record_set
    }


def gen_change_batch(changes=ANY, comment=ANY):
    """Generate expected change batch."""
    return {
        'Comment': comment,
        'Changes': changes
    }


def gen_change_resource_record_sets(**kwargs):
    """Generate response for change_resource_record_sets."""
    data = {
        'Id': 'placeholder_id',
        'Status': 'PENDING',
        'SubmittedAt': datetime.now(),
        'Comment': 'placeholder_comment'
    }
    data.update(kwargs)
    return {'ChangeInfo': data}


def gen_domain_validation_option(**kwargs):
    """Generate a domain validation entry."""
    data = {
        'DomainName': 'place_holder_domain_name',
        'ValidationStatus': 'PENDING_VALIDATION',
        'ValidationDomain': 'place_holder_validation_domain',
        'ResourceRecord': {
            'Name': 'domain_name',
            'Type': 'CNAME',
            'Value': 'record_value'
        },
        'ValidationMethod': 'DNS'
    }
    data.update(kwargs)
    return data


def gen_record_set(use_resource_record=False, **kwargs):
    """Generate a record set."""
    data = {
        'Name': 'placeholder_name',
        'Type': 'CNAME',
        'Value': 'placeholder_value'
    }
    if use_resource_record:
        data['ResourceRecords'] = kwargs.pop('ResourceRecords', [
            {'Value': kwargs.pop('Value', data['Value'])}
        ])
        del data['Value']

    data.update(kwargs)
    return data


def gen_stack_resource(**kwargs):
    """Generate a response to describe_stack_resources."""
    data = {
        'StackName': 'place_holder_stack_name',
        'LogicalResourceId': 'placeholder_logical_resource_id',
        'ResourceType': 'placeholder_resource_type',
        'Timestamp': datetime.now(),
        'ResourceStatus': 'CREATE_IN_PROGRESS'
    }
    data.update(kwargs)
    return data


class TestCertificate(object):
    """Tests for runway.cfngin.hooks.acm.Certificate."""

    def test_attributes(self, cfngin_context):
        """Test attributes set during __init__."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        result = Certificate(context=cfngin_context,
                             provider=MagicMock(),
                             alt_names=['example.net'],
                             domain='example.com',
                             hosted_zone_id='test',
                             stack_name='stack-name',
                             ttl=13)

        assert result.stack_name == 'stack-name'
        assert result.properties.DomainName == 'example.com'
        assert result.properties.SubjectAlternativeNames == ['example.net']
        # value tested in base class; just ensure its not None
        assert result.properties.Tags
        assert result.properties.ValidationMethod == 'DNS'

        # blueprint attributes
        assert result.blueprint.VARIABLES['DomainName']
        assert result.blueprint.VARIABLES['ValidateRecordTTL']

        # template attributes
        template = result.blueprint.template
        assert template.description == result.get_template_description()
        assert not template.metadata
        assert not template.conditions
        assert not template.mappings
        assert template.outputs['DomainName'].Value.to_dict() == \
            {'Ref': 'DomainName'}
        assert template.outputs['ValidateRecordTTL'].Value.to_dict() == \
            {'Ref': 'ValidateRecordTTL'}
        assert not template.parameters
        assert isinstance(template.resources['Certificate'],
                          CertificateResource)
        assert not template.rules
        assert template.version == '2010-09-09'
        assert not template.transform

        # stack attributes
        assert result.stack.fqn == 'test-stack-name'
        assert result.stack._blueprint == result.blueprint

    def test_domain_changed(self, cfngin_context):
        """Test for domain_changed."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        provider = MagicMock()

        cert = Certificate(context=cfngin_context,
                           provider=provider,
                           domain='example.com',
                           hosted_zone_id='test')

        domain_match = {'DomainName': 'example.com'}
        checks = [
            # is_stack_recreatable, is_stack_in_progress, is_stack_rolling_back, get_outputs
            (False, False, False, {'DomainName': 'nope'}),
            (False, False, False, domain_match),
            (False, False, False, {}),
            (False, False, False, StackDoesNotExist('test')),
            (True, False, False, domain_match),
            (False, True, False, domain_match),
            (False, False, True, domain_match)
        ]

        provider.get_stack.return_value = True
        provider.is_stack_recreatable.side_effect = [x[0] for x in checks]
        provider.is_stack_in_progress.side_effect = [x[1] for x in checks]
        provider.is_stack_rolling_back.side_effect = [x[2] for x in checks]
        provider.get_outputs.side_effect = [x[3] for x in checks]

        # output based
        assert cert.domain_changed()  # {'DomainName': 'nope'}
        assert not cert.domain_changed()  # {'DomainName': 'example.com'}
        assert not cert.domain_changed()  # {}
        assert not cert.domain_changed()  # StackDoesNotExist('test')
        # status based
        assert not cert.domain_changed()
        assert not cert.domain_changed()
        assert not cert.domain_changed()

    def test_get_certificate(self, cfngin_context, patch_time):
        """Test get_certificate."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        provider = MagicMock(cloudformation=boto3.client('cloudformation'))
        cfn_stubber = Stubber(provider.cloudformation)

        cert = Certificate(context=cfngin_context,
                           provider=provider,
                           domain='example.com',
                           hosted_zone_id='test')

        expected_request = {'StackName': 'test-example-com',
                            'LogicalResourceId': 'Certificate'}

        cfn_stubber.add_response('describe_stack_resources',
                                 {'StackResources': []},
                                 expected_request)
        cfn_stubber.add_response(
            'describe_stack_resources',
            {'StackResources': [gen_stack_resource()]},
            expected_request
        )
        cfn_stubber.add_response(
            'describe_stack_resources',
            {'StackResources': [
                gen_stack_resource(PhysicalResourceId='success')
            ]},
            expected_request
        )

        with cfn_stubber:
            assert cert.get_certificate() == 'success'

        cfn_stubber.assert_no_pending_responses()

    @pytest.mark.parametrize('status', ['PENDING_VALIDATION', 'SUCCESS', 'FAILED'])
    def test_get_validation_record(self, cfngin_context, monkeypatch,
                                   patch_time, status):
        """Test get_validation_record."""
        # setup context
        acm_stubber = cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected_request = {'CertificateArn': cert_arn}
        validate_option_missing_record = gen_domain_validation_option(
            ValidationStatus=status
        )
        del validate_option_missing_record['ResourceRecord']

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)

        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn),
            expected_request
        )
        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                validate_option_missing_record
                            ]),
            expected_request
        )
        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                gen_domain_validation_option(ValidationStatus=status)
                            ]),
            expected_request
        )

        with acm_stubber:
            assert cert.get_validation_record(status=status) == \
                gen_domain_validation_option()['ResourceRecord']
        acm_stubber.assert_no_pending_responses()

    @pytest.mark.parametrize('check,found', [('PENDING_VALIDATION', 'SUCCESS'),
                                             ('SUCCESS', 'PENDING_VALIDATION'),
                                             ('FAILED', 'SUCCESS')])
    def test_get_validation_record_status_missmatch(self, cfngin_context,
                                                    check, found):
        """Test get get_validation_record with a missmatched record status."""
        # setup context
        acm_stubber = cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected_request = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                gen_domain_validation_option(ValidationStatus=found)
                            ]),
            expected_request
        )

        with acm_stubber, pytest.raises(ValueError) as excinfo:
            cert.get_validation_record(cert_arn=cert_arn, status=check)

        assert 'No validations with status' in str(excinfo.value)
        acm_stubber.assert_no_pending_responses()

    def test_get_validation_record_gt_one(self, cfngin_context):
        """Test get get_validation_record more than one result."""
        # setup context
        acm_stubber = cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected_request = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                gen_domain_validation_option(),
                                gen_domain_validation_option()
                            ]),
            expected_request
        )

        with acm_stubber, pytest.raises(ValueError) as excinfo:
            cert.get_validation_record(cert_arn=cert_arn)

        assert 'only one option is supported' in str(excinfo.value)
        acm_stubber.assert_no_pending_responses()

    def test_put_record_set(self, cfngin_context):
        """Test put_record."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        r53_stubber = cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        r53_stubber.add_response(
            'change_resource_record_sets',
            gen_change_resource_record_sets(),
            {
                'HostedZoneId': cert.args.hosted_zone_id,
                'ChangeBatch': gen_change_batch(changes=[
                    gen_change(record_set=gen_record_set(use_resource_record=True,
                                                         TTL=cert.args.ttl))
                ])
            }
        )

        with r53_stubber:
            assert not cert.put_record_set(gen_record_set())
        r53_stubber.assert_no_pending_responses()

    def test_remove_validation_records(self, cfngin_context, monkeypatch):
        """Test remove_validation_records."""
        # setup context
        acm_stubber = cfngin_context.add_stubber('acm', 'us-east-1')
        r53_stubber = cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected_cert_request = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)

        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                gen_domain_validation_option()
                            ]),
            expected_cert_request
        )
        acm_stubber.add_response(
            'describe_certificate',
            gen_certificate(CertificateArn=cert_arn,
                            DomainValidationOptions=[
                                gen_domain_validation_option(ValidationMethod='EMAIL')
                            ]),
            expected_cert_request
        )

        r53_stubber.add_response(
            'change_resource_record_sets',
            gen_change_resource_record_sets(),
            {
                'HostedZoneId': cert.args.hosted_zone_id,
                'ChangeBatch': gen_change_batch(changes=[
                    gen_change(action='DELETE',
                               record_set=gen_record_set(
                                   use_resource_record=True,
                                   TTL=cert.args.ttl,
                                   **gen_domain_validation_option()['ResourceRecord']
                               ))
                ])
            }
        )

        with acm_stubber, r53_stubber, pytest.raises(ValueError) as excinfo:
            assert not cert.remove_validation_records()
            cert.remove_validation_records()

        acm_stubber.assert_no_pending_responses()
        r53_stubber.assert_no_pending_responses()
        assert str(excinfo.value) == 'Must provide one of more record sets'

    def test_update_record_set(self, cfngin_context):
        """Test update_record_set."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        r53_stubber = cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        r53_stubber.add_response(
            'change_resource_record_sets',
            gen_change_resource_record_sets(),
            {
                'HostedZoneId': cert.args.hosted_zone_id,
                'ChangeBatch': gen_change_batch(changes=[
                    gen_change(action='UPSERT',
                               record_set=gen_record_set(use_resource_record=True,
                                                         TTL=cert.args.ttl))
                ])
            }
        )

        with r53_stubber:
            assert not cert.update_record_set(gen_record_set())
        r53_stubber.assert_no_pending_responses()

    def test_deploy(self, cfngin_context, monkeypatch):
        """Test deploy."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack', lambda: STATUS.new)
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)
        monkeypatch.setattr(cert, 'get_validation_record',
                            lambda x: 'get_validation_record'
                            if x == cert_arn else ValueError)
        monkeypatch.setattr(cert, 'put_record_set',
                            lambda x: None
                            if x == 'get_validation_record' else ValueError)
        monkeypatch.setattr(cert, '_wait_for_stack',
                            lambda x, last_status: None)

        assert cert.deploy() == expected

    def test_deploy_update(self, cfngin_context, monkeypatch):
        """Test deploy update stack."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack', lambda: STATUS.update)
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)
        monkeypatch.setattr(cert, 'get_validation_record',
                            lambda x, status: 'get_validation_record'
                            if x == cert_arn and status == 'SUCCESS' else ValueError)
        monkeypatch.setattr(cert, 'update_record_set',
                            lambda x: None
                            if x == 'get_validation_record' else ValueError)
        monkeypatch.setattr(cert, '_wait_for_stack',
                            lambda x, last_status: None)

        assert cert.deploy() == expected

    def test_deploy_no_change(self, cfngin_context, monkeypatch):
        """Test deploy no change."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack', lambda: STATUS.no)
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)

        assert cert.deploy() == expected

    def test_deploy_recreate(self, cfngin_context, monkeypatch):
        """Test deploy with stack recreation."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'
        expected = {'CertificateArn': cert_arn}

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack', lambda: STATUS.recreate)
        monkeypatch.setattr(cert, 'get_certificate',
                            MagicMock(side_effect=['old', cert_arn]))
        monkeypatch.setattr(cert, '_wait_for_stack',
                            MagicMock(side_effect=[STATUS.new, None]))
        monkeypatch.setattr(cert, 'get_validation_record',
                            lambda x: 'get_validation_record'
                            if x == cert_arn else ValueError)
        monkeypatch.setattr(cert, 'put_record_set',
                            lambda x: None
                            if x == 'get_validation_record' else ValueError)

        assert cert.deploy() == expected

    def test_deploy_domain_changed(self, cfngin_context, monkeypatch):
        """Test deploy domain changed."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: True)

        assert not cert.deploy()

    def test_deploy_error_destroy(self, cfngin_context, monkeypatch):
        """Test deploy with errors that result in destroy being called."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert_arn = 'arn:aws:acm:us-east-1:012345678901:certificate/test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack', lambda: STATUS.new)
        monkeypatch.setattr(cert, 'get_certificate', lambda: cert_arn)
        monkeypatch.setattr(cert, 'get_validation_record',
                            lambda x: 'get_validation_record'
                            if x == cert_arn else ValueError)
        monkeypatch.setattr(cert, 'put_record_set',
                            MagicMock(side_effect=[
                                cert.r53_client.exceptions.InvalidChangeBatch({}, ''),
                                cert.r53_client.exceptions.NoSuchHostedZone({}, ''),
                                None
                            ]))
        monkeypatch.setattr(cert, 'destroy',
                            lambda records, skip_r53: check_bool_is_true(skip_r53))
        monkeypatch.setattr(cert, '_wait_for_stack',
                            MagicMock(side_effect=StackFailed('test')))

        assert not cert.deploy()  # cert.r53_client.exceptions.InvalidChangeBatch
        assert not cert.deploy()  # cert.r53_client.exceptions.NoSuchHostedZone

        monkeypatch.setattr(cert, 'destroy',
                            lambda records, skip_r53: check_bool_is_false(skip_r53))
        assert not cert.deploy()  # StackFailed

    def test_deploy_error_no_destroy(self, cfngin_context, monkeypatch):
        """Test deploy with errors that don't result in destroy being called."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'domain_changed', lambda: False)
        monkeypatch.setattr(cert, 'deploy_stack',
                            MagicMock(side_effect=StackUpdateBadStatus('test',
                                                                       'test',
                                                                       'test')))

        assert not cert.deploy()

    def test_destory(self, cfngin_context, monkeypatch):
        """Test destory."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        # should only be called once
        monkeypatch.setattr(cert, 'remove_validation_records',
                            MagicMock(return_value=None))
        monkeypatch.setattr(cert, 'destroy_stack', lambda wait: None)

        assert cert.destroy()
        assert cert.destroy(skip_r53=True)
        assert cert.remove_validation_records.call_count == 1  # pylint: disable=no-member

    def test_destory_aws_errors(self, cfngin_context, monkeypatch):
        """Test destory with errors from AWS."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')

        monkeypatch.setattr(
            cert,
            'remove_validation_records',
            MagicMock(side_effect=[
                cert.r53_client.exceptions.InvalidChangeBatch({}, ''),
                cert.r53_client.exceptions.NoSuchHostedZone({}, ''),
                cert.acm_client.exceptions.ResourceNotFoundException({}, '')
            ])
        )
        monkeypatch.setattr(cert, 'destroy_stack', lambda wait: None)

        assert cert.destroy()
        assert cert.destroy()
        assert cert.destroy()

    def test_destroy_raise_client_error(self, cfngin_context, monkeypatch):
        """Test destory with ClientError raised."""
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        def build_client_error(msg):
            """Raise a ClientError."""
            return ClientError({'Error': {'Message': msg}}, 'test')

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'destroy_stack', lambda wait: None)

        def raise_stack_not_exist(_records):
            """Raise ClientError mimicing stack not existing."""
            raise build_client_error('Stack with id {} does not exist'.format(
                cert.stack.fqn
            ))

        def raise_other(_records):
            """Raise other ClientError."""
            raise build_client_error('something')

        monkeypatch.setattr(cert, 'remove_validation_records',
                            raise_stack_not_exist)
        assert cert.destroy()

        monkeypatch.setattr(cert, 'remove_validation_records',
                            raise_other)
        with pytest.raises(ClientError) as excinfo:
            cert.destroy()
        assert 'something' in str(excinfo.value)

    @pytest.mark.parametrize('stage,expected', [('post_deploy', 'deploy'),
                                                ('post_destroy', 'destroy'),
                                                ('pre_deploy', 'deploy'),
                                                ('pre_destroy', 'destroy')])
    def test_stage_methods(self, cfngin_context, monkeypatch, stage, expected):
        """Test stage methods.

        All of these call a different method that is being tested separately
        so this test just ensures each stage is mapped to the correct method
        on the backend.

        """
        # setup context
        cfngin_context.add_stubber('acm', 'us-east-1')
        cfngin_context.add_stubber('route53', 'us-east-1')
        cfngin_context.config.namespace = 'test'

        cert = Certificate(context=cfngin_context,
                           provider=MagicMock(),
                           domain='example.com',
                           hosted_zone_id='test')
        monkeypatch.setattr(cert, 'deploy', lambda: 'deploy')
        monkeypatch.setattr(cert, 'destroy', lambda: 'destroy')

        assert getattr(cert, stage)() == expected
